# Copyright 2018 ASI Data Science
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import time

import boto3

import sherlockml


class SherlockMLDatasetsError(Exception):
    pass


def project_id_from_environment():
    try:
        project_id = os.environ["SHERLOCKML_PROJECT_ID"]
    except KeyError:
        raise SherlockMLDatasetsError(
            "No SHERLOCKML_PROJECT_ID in environment - set the project ID "
            "explicitly to use outside of SherlockML"
        )
    return project_id


SECRETS_CACHE_TTL = 10


class DatasetsSession(object):
    def __init__(self):
        self.secret_client = sherlockml.client("secret")
        self.bucket_cache = {}
        self.secrets_cache = {}

    def bucket(self, project_id):
        if project_id is None:
            project_id = project_id_from_environment()
        if project_id not in self.bucket_cache:
            secrets = self.secret_client.datasets_secrets(project_id)
            self.bucket_cache[project_id] = secrets.bucket
        return self.bucket_cache[project_id]

    def _get_verified_secrets(self, project_id):
        secrets = self.secret_client.datasets_secrets(project_id)
        tries = 1
        while not secrets.verified:
            if tries >= 30:
                raise ValueError("Secrets not verified after 60 seconds")
            time.sleep(2)
            secrets = self.secret_client.datasets_secrets(project_id)
            tries += 1
        return secrets

    def _cached_secrets(self, project_id):
        if (
            project_id not in self.secrets_cache
            or self.secrets_cache[project_id][1] + SECRETS_CACHE_TTL
            < time.time()
        ):
            secrets = self._get_verified_secrets(project_id)
            self.secrets_cache[project_id] = secrets, time.time()
        return self.secrets_cache[project_id][0]

    def s3_client(self, project_id=None):

        if project_id is None:
            project_id = project_id_from_environment()

        secrets = self._cached_secrets(project_id)

        boto_session = boto3.session.Session(
            aws_access_key_id=secrets.access_key,
            aws_secret_access_key=secrets.secret_key,
            region_name="eu-west-1",
        )

        return boto_session.client("s3")


DATASETS_SESSION = None


def get():
    global DATASETS_SESSION
    if DATASETS_SESSION is None:
        DATASETS_SESSION = DatasetsSession()
    return DATASETS_SESSION



# Copyright 2018 ASI Data Science
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from datetime import datetime
from dateutil.tz import UTC
from uuid import uuid4

from marshmallow import ValidationError

from sherlockml.clients.workspace import (
    Directory,
    File,
    FileNodeSchema,
    ListResponse,
    ListResponseSchema,
    WorkspaceClient,
)
from tests.clients.fixtures import PROFILE


PROJECT_ID = uuid4()

FILE = File(
    path="/path/to/test-directory/test-file",
    name="test-file",
    last_modified=datetime(2017, 8, 9, 14, 21, 13, tzinfo=UTC),
    size=906,
)

FILE_BODY = {
    "path": FILE.path,
    "name": FILE.name,
    "last_modified": "2017-08-09T14:21:13Z",
    "size": FILE.size,
    "type": "file",
}

DIRECTORY = Directory(
    path="/path/to/test-directory",
    name="test-directory",
    last_modified=datetime(2018, 1, 15, 15, 18, 45, tzinfo=UTC),
    size=6144,
    truncated=False,
    content=[FILE],
)

DIRECTORY_BODY = {
    "path": DIRECTORY.path,
    "name": DIRECTORY.name,
    "last_modified": "2018-01-15T15:18:45Z",
    "size": DIRECTORY.size,
    "type": "directory",
    "truncated": DIRECTORY.truncated,
    "content": [FILE_BODY],
}

LIST_RESPONSE = ListResponse(
    project_id=uuid4(), path="/path/to/test-directory/", content=[DIRECTORY]
)

LIST_RESPONSE_BODY = {
    "project_id": str(LIST_RESPONSE.project_id),
    "path": LIST_RESPONSE.path,
    "content": [DIRECTORY_BODY],
}


def test_file_node_schema_file():
    assert FileNodeSchema().load(FILE_BODY) == FILE


def test_file_node_schema_directory():
    assert FileNodeSchema().load(DIRECTORY_BODY) == DIRECTORY


def test_file_node_schema_invalid():
    with pytest.raises(ValidationError):
        FileNodeSchema().load({})


def test_list_response_schema():
    assert ListResponseSchema().load(LIST_RESPONSE_BODY) == LIST_RESPONSE


def test_list_response_schema_invalid():
    with pytest.raises(ValidationError):
        ListResponseSchema().load({})


def test_workspace_client_get(mocker):
    mocker.patch.object(WorkspaceClient, "_get", return_value=LIST_RESPONSE)
    schema_mock = mocker.patch(
        "sherlockml.clients.workspace.ListResponseSchema"
    )

    client = WorkspaceClient(PROFILE)
    assert client.list(
        PROJECT_ID, prefix="/path/to/test-directory/", depth=1
    ) == [DIRECTORY]

    schema_mock.assert_called_once_with()
    WorkspaceClient._get.assert_called_once_with(
        "/project/{}/file".format(PROJECT_ID),
        schema_mock.return_value,
        params={"prefix": "/path/to/test-directory/", "depth": 1},
    )



# Copyright 2018 ASI Data Science
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from collections import namedtuple

from marshmallow import Schema, fields, post_load

from sherlockml.clients.base import BaseClient


NodeType = namedtuple(
    "NodeType",
    [
        "id",
        "name",
        "instance_group",
        "max_interactive_instances",
        "max_job_instances",
        "milli_cpus",
        "memory_mb",
        "num_gpus",
        "gpu_name",
        "cost_usd_per_hour",
    ],
)


class NodeTypeSchema(Schema):

    id = fields.String(data_key="nodeTypeId", required=True)
    name = fields.String(missing=None)
    instance_group = fields.String(data_key="instanceGroup", missing=None)
    max_interactive_instances = fields.Integer(
        data_key="maxInteractiveInstances", required=True
    )
    max_job_instances = fields.Integer(
        data_key="maxJobInstances", required=True
    )
    milli_cpus = fields.Integer(data_key="milliCpus", required=True)
    memory_mb = fields.Integer(data_key="memoryMb", required=True)
    num_gpus = fields.Integer(data_key="numGpus", required=True)
    gpu_name = fields.String(data_key="gpuName", missing=None)
    cost_usd_per_hour = fields.Decimal(
        data_key="costUsdPerHour", required=True
    )

    @post_load
    def make_node_type(self, data):
        return NodeType(**data)


class ClusterClient(BaseClient):

    SERVICE_NAME = "klostermann"

    def list_single_tenanted_node_types(
        self,
        interactive_instances_configured=None,
        job_instances_configured=None,
    ):
        """Get information on single tenanted node types from the cluster.

        Parameters
        ----------
        interactive_instances_configured : bool, optional
            If True, only get node types which are configured to support
            interactive instances, or if False, those which are not configured
            to support interactive instances
        job_instances_configured : bool, optional
            If True, only get node types which are configured to support job
            instances, or if False, those which are not configured to support
            job instances

        Returns
        -------
        list of NodeType
        """

        query_params = {}
        if interactive_instances_configured is not None:
            query_params["interactiveInstancesConfigured"] = (
                "true" if interactive_instances_configured else "false"
            )
        if job_instances_configured is not None:
            query_params["jobInstancesConfigured"] = (
                "true" if job_instances_configured else "false"
            )

        return self._get(
            "/node-type/single-tenanted",
            NodeTypeSchema(many=True),
            params=query_params,
        )

    def configure_single_tenanted_node_type(
        self,
        node_type_id,
        name,
        instance_group,
        max_interactive_instances,
        max_job_instances,
    ):
        payload = {
            "name": name,
            "instanceGroup": instance_group,
            "maxInteractiveInstances": max_interactive_instances,
            "maxJobInstances": max_job_instances,
        }
        return self._put_raw(
            "/node-type/single-tenanted/{}/configuration".format(node_type_id),
            json=payload,
        )

    def disable_single_tenanted_node_type(self, node_type_id):
        return self._delete_raw(
            "/node-type/single-tenanted/{}/configuration".format(node_type_id)
        )



# Copyright 2018 ASI Data Science
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from collections import namedtuple

from marshmallow import Schema, fields, post_load

from sherlockml.clients.base import BaseClient


Project = namedtuple("Project", ["id", "name", "owner_id"])


class ProjectSchema(Schema):

    id = fields.UUID(data_key="projectId", required=True)
    name = fields.Str(required=True)
    owner_id = fields.UUID(data_key="ownerId", required=True)

    @post_load
    def make_project(self, data):
        return Project(**data)


class ProjectClient(BaseClient):

    SERVICE_NAME = "casebook"

    def create(self, owner_id, project_name):
        payload = {"owner_id": str(owner_id), "name": project_name}
        return self._post("/project", ProjectSchema(), json=payload)

    def get(self, project_id):
        endpoint = "/project/{}".format(project_id)
        return self._get(endpoint, ProjectSchema())

    def get_by_owner_and_name(self, owner_id, project_name):
        endpoint = "/project/{}/{}".format(owner_id, project_name)
        return self._get(endpoint, ProjectSchema())

    def list_accessible_by_user(self, user_id):
        endpoint = "/user/{}".format(user_id)
        return self._get(endpoint, ProjectSchema(many=True))



# Copyright 2018 ASI Data Science
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import requests
import marshmallow
from six.moves import urllib

from sherlockml.clients.auth import SherlockMLAuth


class BadResponseStatus(Exception):
    def __init__(self, response):
        self.response = response


class Unauthorized(BadResponseStatus):
    pass


class NotFound(BadResponseStatus):
    pass


class InvalidResponse(Exception):
    pass


def _service_url(profile, service, endpoint=""):
    host = "{}.{}".format(service, profile.domain)
    url_parts = (profile.protocol, host, endpoint, None, None)
    return urllib.parse.urlunsplit(url_parts)


def _check_status(response):
    if response.status_code == 401:
        raise Unauthorized(response)
    elif response.status_code == 404:
        raise NotFound(response)
    elif response.status_code >= 400:
        raise BadResponseStatus(response)


def _deserialise_response(schema, response):
    try:
        response_json = response.json()
    except ValueError:
        raise InvalidResponse("response body was not valid JSON")

    try:
        data = schema.load(response_json)
    except marshmallow.ValidationError:
        # TODO: log validation errors and possibly include them in raised
        # exception
        raise InvalidResponse("response content did not match expected format")

    return data


class BaseClient(object):

    SERVICE_NAME = None

    def __init__(self, profile):
        if self.SERVICE_NAME is None:
            raise RuntimeError(
                "must set SERVICE_NAME in subclasses of BaseClient"
            )
        self.profile = profile
        self._http_session_cache = None

    @property
    def http_session(self):
        if self._http_session_cache is None:
            self._http_session_cache = requests.Session()
            self._http_session_cache.auth = SherlockMLAuth(
                _service_url(self.profile, "hudson"),
                self.profile.client_id,
                self.profile.client_secret,
            )
        return self._http_session_cache

    def _request(self, method, endpoint, check_status=True, *args, **kwargs):
        url = _service_url(self.profile, self.SERVICE_NAME, endpoint)
        response = self.http_session.request(method, url, *args, **kwargs)
        if check_status:
            _check_status(response)
        return response

    def _get_raw(self, endpoint, *args, **kwargs):
        return self._request("GET", endpoint, *args, **kwargs)

    def _get(self, endpoint, schema, **kwargs):
        response = self._get_raw(endpoint, **kwargs)
        return _deserialise_response(schema, response)

    def _post_raw(self, endpoint, *args, **kwargs):
        return self._request("POST", endpoint, *args, **kwargs)

    def _post(self, endpoint, schema, **kwargs):
        response = self._post_raw(endpoint, **kwargs)
        return _deserialise_response(schema, response)

    def _put_raw(self, endpoint, *args, **kwargs):
        return self._request("PUT", endpoint, *args, **kwargs)

    def _put(self, endpoint, schema, **kwargs):
        response = self._put_raw(endpoint, **kwargs)
        return _deserialise_response(schema, response)

    def _delete_raw(self, endpoint, *args, **kwargs):
        return self._request("DELETE", endpoint, *args, **kwargs)

    def _delete(self, endpoint, schema, **kwargs):
        response = self._delete_raw(endpoint, **kwargs)
        return _deserialise_response(schema, response)



# Copyright 2018 ASI Data Science
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import sherlockml


def test_client(mocker):
    profile_mock = mocker.patch("sherlockml.config.resolve_profile")
    for_resource_mock = mocker.patch("sherlockml.clients.for_resource")

    config_overrides = {"foo": "bar"}
    sherlockml.client("test-resource", **config_overrides)

    profile_mock.assert_called_once_with(**config_overrides)
    for_resource_mock.assert_called_once_with("test-resource")

    returned_profile = profile_mock.return_value
    returned_class = for_resource_mock.return_value
    returned_class.assert_called_once_with(returned_profile)



import os

import pytest

from sherlockml import config


SAMPLE_CONFIG_CONTENT = """
[default]
domain = test.domain.com
protocol = test-protocol
client_id = test-client-id
client_secret = test-client-secret

[empty profile]
"""

DEFAULT_PROFILE = config.Profile(
    domain="test.domain.com",
    protocol="test-protocol",
    client_id="test-client-id",
    client_secret="test-client-secret",
)
OTHER_PROFILE = config.Profile(
    domain="other.domain.com",
    protocol="other-protocol",
    client_id="other-client-id",
    client_secret="other-client-secret",
)
CREDENTIALS_ONLY_PROFILE = config.Profile(
    domain=None,
    protocol=None,
    client_id="test-client-id",
    client_secret="test-client-secret",
)
PROFILE_WITHOUT_ID = config.Profile(
    domain="test.domain.com",
    protocol="test-protocol",
    client_id=None,
    client_secret="test-client-secret",
)
PROFILE_WITHOUT_SECRET = config.Profile(
    domain="test.domain.com",
    protocol="test-protocol",
    client_id="test-client-id",
    client_secret=None,
)
EMPTY_PROFILE = config.Profile(
    domain=None, protocol=None, client_id=None, client_secret=None
)

SAMPLE_CONFIG = {"default": DEFAULT_PROFILE, "empty profile": EMPTY_PROFILE}


def test_load(tmpdir):
    file = tmpdir.join("config")
    file.write(SAMPLE_CONFIG_CONTENT)
    assert config.load(file) == SAMPLE_CONFIG


def test_load_missing():
    assert config.load("does-not-exist") == {}


@pytest.mark.parametrize(
    "profile_name, expected_profile",
    [("default", DEFAULT_PROFILE), ("missing profile", EMPTY_PROFILE)],
)
def test_load_profile(mocker, profile_name, expected_profile):
    mocker.patch("sherlockml.config.load", return_value=SAMPLE_CONFIG)

    assert config.load_profile("test/path", profile_name) == expected_profile

    config.load.assert_called_once_with("test/path")


def test_default_credentials_path(mocker):
    mocker.patch.dict(os.environ, {"HOME": "/foo/bar"})
    expected_path = "/foo/bar/.config/sherlockml/credentials"
    assert config._default_credentials_path() == expected_path


def test_default_credentials_path_xdg_home(mocker):
    mocker.patch.dict(os.environ, {"XDG_CONFIG_HOME": "/xdg/home"})
    expected_path = "/xdg/home/sherlockml/credentials"
    assert config._default_credentials_path() == expected_path


def test_resolve_profile(mocker):
    mocker.patch(
        "sherlockml.config.load_profile", return_value=DEFAULT_PROFILE
    )
    mocker.patch("sherlockml.config._default_credentials_path")

    assert config.resolve_profile() == DEFAULT_PROFILE

    config.load_profile.assert_called_once_with(
        config._default_credentials_path.return_value, "default"
    )


def test_resolve_profile_credentials_path_override(mocker):
    mocker.patch(
        "sherlockml.config.load_profile", return_value=DEFAULT_PROFILE
    )

    profile = config.resolve_profile(credentials_path="test/path")
    assert profile == DEFAULT_PROFILE

    config.load_profile.assert_called_once_with("test/path", "default")


def test_resolve_profile_credentials_path_env(mocker):
    mocker.patch(
        "sherlockml.config.load_profile", return_value=DEFAULT_PROFILE
    )
    path = "/path/to/credentials"
    mocker.patch.dict(os.environ, {"SHERLOCKML_CREDENTIALS_PATH": path})

    assert config.resolve_profile() == DEFAULT_PROFILE

    config.load_profile.assert_called_once_with(path, "default")


def test_resolve_profile_profile_name_override(mocker):
    mocker.patch("sherlockml.config.load_profile", return_value=OTHER_PROFILE)
    mocker.patch("sherlockml.config._default_credentials_path")

    profile = config.resolve_profile(profile_name="other")
    assert profile == OTHER_PROFILE

    config.load_profile.assert_called_once_with(
        config._default_credentials_path.return_value, "other"
    )


def test_resolve_profile_profile_name_env(mocker):
    mocker.patch("sherlockml.config.load_profile", return_value=OTHER_PROFILE)
    mocker.patch("sherlockml.config._default_credentials_path")
    mocker.patch.dict(os.environ, {"SHERLOCKML_PROFILE": "other"})

    assert config.resolve_profile() == OTHER_PROFILE

    config.load_profile.assert_called_once_with(
        config._default_credentials_path.return_value, "other"
    )


def test_resolve_profile_overrides(mocker):
    mocker.patch(
        "sherlockml.config.load_profile", return_value=DEFAULT_PROFILE
    )
    profile = config.resolve_profile(
        domain="other.domain.com",
        protocol="other-protocol",
        client_id="other-client-id",
        client_secret="other-client-secret",
    )
    assert profile == OTHER_PROFILE


def test_resolve_profile_env(mocker):
    mocker.patch(
        "sherlockml.config.load_profile", return_value=DEFAULT_PROFILE
    )
    mocker.patch.dict(
        os.environ,
        {
            "SHERLOCKML_DOMAIN": "other.domain.com",
            "SHERLOCKML_PROTOCOL": "other-protocol",
            "SHERLOCKML_CLIENT_ID": "other-client-id",
            "SHERLOCKML_CLIENT_SECRET": "other-client-secret",
        },
    )
    assert config.resolve_profile() == OTHER_PROFILE


def test_resolve_profile_defaults(mocker):
    mocker.patch(
        "sherlockml.config.load_profile", return_value=CREDENTIALS_ONLY_PROFILE
    )
    profile = config.resolve_profile()
    assert profile == config.Profile(
        domain=config.DEFAULT_DOMAIN,
        protocol=config.DEFAULT_PROTOCOL,
        client_id="test-client-id",
        client_secret="test-client-secret",
    )


def test_resolve_profile_missing_client_id(mocker):
    mocker.patch(
        "sherlockml.config.load_profile", return_value=PROFILE_WITHOUT_ID
    )
    with pytest.raises(config.CredentialsError, match="client_id"):
        config.resolve_profile()


def test_resolve_profile_missing_client_secret(mocker):
    mocker.patch(
        "sherlockml.config.load_profile", return_value=PROFILE_WITHOUT_SECRET
    )
    with pytest.raises(config.CredentialsError, match="client_secret"):
        config.resolve_profile()



# Copyright 2018 ASI Data Science
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from collections import namedtuple

from marshmallow import Schema, fields, post_load

from sherlockml.clients.base import BaseClient


User = namedtuple("User", ["id"])


class UserSchema(Schema):
    id = fields.UUID(data_key="userId", required=True)

    @post_load
    def make_user(self, data):
        return User(**data)


AuthenticationResponse = namedtuple("AuthenticationResponse", ["user"])


class AuthenticationResponseSchema(Schema):
    user = fields.Nested(UserSchema, data_key="account", required=True)

    @post_load
    def make_authentication_response(self, data):
        return AuthenticationResponse(**data)


class UserClient(BaseClient):

    SERVICE_NAME = "hudson"

    def authenticated_user_id(self):
        data = self._get("/authenticate", AuthenticationResponseSchema())
        return data.user.id



# Copyright 2018 ASI Data Science
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import uuid
from datetime import datetime

import pytest
from marshmallow import ValidationError
from dateutil.tz import UTC

from sherlockml.clients.server import (
    Server,
    ServerStatus,
    Service,
    ServerSchema,
    ServiceSchema,
    ServerClient,
    ServerIdSchema,
    SharedServerResources,
    DedicatedServerResources,
)
from tests.clients.fixtures import PROFILE


SERVICE = Service(
    name="hound",
    host="server-99999-hound.domain.com",
    port=443,
    scheme="https",
    uri="https://server-99999-hound.domain.com:443",
)

SERVICE_BODY = {
    "name": SERVICE.name,
    "host": SERVICE.host,
    "port": SERVICE.port,
    "scheme": SERVICE.scheme,
    "uri": SERVICE.uri,
}

PROJECT_ID = uuid.uuid4()
OWNER_ID = uuid.uuid4()
SERVER_ID = uuid.uuid4()
CREATED_AT = datetime(2018, 3, 10, 11, 32, 6, 247000, tzinfo=UTC)
CREATED_AT_STRING = "2018-03-10T11:32:06.247Z"

SHARED_RESOURCES = SharedServerResources(milli_cpus=1000, memory_mb=4000)
SHARED_SERVER = Server(
    id=SERVER_ID,
    project_id=PROJECT_ID,
    owner_id=OWNER_ID,
    name="test server",
    type="jupyter",
    resources=SHARED_RESOURCES,
    created_at=CREATED_AT,
    status=ServerStatus.RUNNING,
    services=[SERVICE],
)
SHARED_SERVER_BODY = {
    "instanceId": str(SERVER_ID),
    "projectId": str(PROJECT_ID),
    "ownerId": str(OWNER_ID),
    "name": SHARED_SERVER.name,
    "instanceType": SHARED_SERVER.type,
    "instanceSizeType": "custom",
    "instanceSize": {
        "milliCpus": SHARED_RESOURCES.milli_cpus,
        "memoryMb": SHARED_RESOURCES.memory_mb,
    },
    "createdAt": CREATED_AT_STRING,
    "status": "running",
    "services": [SERVICE_BODY],
}

DEDICATED_RESOURCES = DedicatedServerResources(node_type="m4.xlarge")
DEDICATED_SERVER = Server(
    id=SERVER_ID,
    project_id=PROJECT_ID,
    owner_id=OWNER_ID,
    name="test server",
    type="jupyter",
    resources=DEDICATED_RESOURCES,
    created_at=CREATED_AT,
    status=ServerStatus.RUNNING,
    services=[SERVICE],
)
DEDICATED_SERVER_BODY = {
    "instanceId": str(SERVER_ID),
    "projectId": str(PROJECT_ID),
    "ownerId": str(OWNER_ID),
    "name": DEDICATED_SERVER.name,
    "instanceType": DEDICATED_SERVER.type,
    "instanceSizeType": DEDICATED_RESOURCES.node_type,
    "createdAt": CREATED_AT_STRING,
    "status": "running",
    "services": [SERVICE_BODY],
}

SERVER_ID_BODY = {"instanceId": str(SERVER_ID)}


def test_service_schema():
    data = ServiceSchema().load(SERVICE_BODY)
    assert data == SERVICE


def test_service_schema_invalid():
    with pytest.raises(ValidationError):
        ServiceSchema().load({})


@pytest.mark.parametrize(
    "body, expected",
    [
        (SHARED_SERVER_BODY, SHARED_SERVER),
        (DEDICATED_SERVER_BODY, DEDICATED_SERVER),
    ],
)
def test_server_schema(body, expected):
    data = ServerSchema().load(body)
    assert data == expected


def test_server_schema_invalid():
    with pytest.raises(ValidationError):
        ServerSchema().load({})


def test_server_schema_invalid_missing_instance_size():
    body = SHARED_SERVER_BODY.copy()
    del body["instanceSize"]
    with pytest.raises(ValidationError):
        ServerSchema().load(body)


def test_server_id_schema():
    data = ServerIdSchema().load(SERVER_ID_BODY)
    assert data == SERVER_ID


def test_server_id_schema_invalid():
    with pytest.raises(ValidationError):
        ServerIdSchema().load({})


def test_server_client_create_shared(mocker):
    mocker.patch.object(ServerClient, "_post", return_value=SERVER_ID)
    schema_mock = mocker.patch("sherlockml.clients.server.ServerIdSchema")

    client = ServerClient(PROFILE)

    server_type = "jupyter"
    name = "test server"
    image_version = "version"
    initial_environment_ids = [uuid.uuid4(), uuid.uuid4()]

    assert (
        client.create(
            PROJECT_ID,
            server_type,
            SHARED_RESOURCES,
            name,
            image_version,
            initial_environment_ids,
        )
        == SERVER_ID
    )

    schema_mock.assert_called_once_with()
    ServerClient._post.assert_called_once_with(
        "/instance/{}".format(PROJECT_ID),
        schema_mock.return_value,
        json={
            "instanceType": server_type,
            "instanceSizeType": "custom",
            "instanceSize": {
                "milliCpus": SHARED_RESOURCES.milli_cpus,
                "memoryMb": SHARED_RESOURCES.memory_mb,
            },
            "name": name,
            "typeVersion": image_version,
            "environmentIds": [
                str(env_id) for env_id in initial_environment_ids
            ],
        },
    )


def test_server_client_create_dedicated(mocker):
    mocker.patch.object(ServerClient, "_post", return_value=SERVER_ID)
    schema_mock = mocker.patch("sherlockml.clients.server.ServerIdSchema")

    client = ServerClient(PROFILE)

    server_type = "jupyter"
    name = "test server"
    image_version = "version"
    initial_environment_ids = [uuid.uuid4(), uuid.uuid4()]

    assert (
        client.create(
            PROJECT_ID,
            server_type,
            DEDICATED_RESOURCES,
            name,
            image_version,
            initial_environment_ids,
        )
        == SERVER_ID
    )

    schema_mock.assert_called_once_with()
    ServerClient._post.assert_called_once_with(
        "/instance/{}".format(PROJECT_ID),
        schema_mock.return_value,
        json={
            "instanceType": server_type,
            "instanceSizeType": DEDICATED_RESOURCES.node_type,
            "name": name,
            "typeVersion": image_version,
            "environmentIds": [
                str(env_id) for env_id in initial_environment_ids
            ],
        },
    )


def test_server_client_create_minimal(mocker):
    mocker.patch.object(ServerClient, "_post", return_value=SERVER_ID)
    schema_mock = mocker.patch("sherlockml.clients.server.ServerIdSchema")

    client = ServerClient(PROFILE)

    server_type = "jupyter"

    assert (
        client.create(PROJECT_ID, server_type, SHARED_RESOURCES) == SERVER_ID
    )

    schema_mock.assert_called_once_with()
    ServerClient._post.assert_called_once_with(
        "/instance/{}".format(PROJECT_ID),
        schema_mock.return_value,
        json={
            "instanceType": server_type,
            "instanceSizeType": "custom",
            "instanceSize": {
                "milliCpus": SHARED_RESOURCES.milli_cpus,
                "memoryMb": SHARED_RESOURCES.memory_mb,
            },
        },
    )


def test_server_client_get(mocker):
    mocker.patch.object(ServerClient, "_get", return_value=SHARED_SERVER)
    schema_mock = mocker.patch("sherlockml.clients.server.ServerSchema")

    client = ServerClient(PROFILE)

    assert client.get(PROJECT_ID, SERVER_ID) == SHARED_SERVER

    schema_mock.assert_called_once_with()
    ServerClient._get.assert_called_once_with(
        "/instance/{}/{}".format(PROJECT_ID, SERVER_ID),
        schema_mock.return_value,
    )


def test_server_client_list(mocker):
    mocker.patch.object(ServerClient, "_get", return_value=[SHARED_SERVER])
    schema_mock = mocker.patch("sherlockml.clients.server.ServerSchema")

    client = ServerClient(PROFILE)

    assert client.list(PROJECT_ID) == [SHARED_SERVER]

    schema_mock.assert_called_once_with(many=True)
    ServerClient._get.assert_called_once_with(
        "/instance/{}".format(PROJECT_ID),
        schema_mock.return_value,
        params=None,
    )


def test_server_client_list_filter_name(mocker):
    mocker.patch.object(ServerClient, "_get", return_value=[SHARED_SERVER])
    schema_mock = mocker.patch("sherlockml.clients.server.ServerSchema")

    client = ServerClient(PROFILE)

    assert client.list(PROJECT_ID, name="foo") == [SHARED_SERVER]

    schema_mock.assert_called_once_with(many=True)
    ServerClient._get.assert_called_once_with(
        "/instance/{}".format(PROJECT_ID),
        schema_mock.return_value,
        params={"name": "foo"},
    )



# Copyright 2018 ASI Data Science
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import posixpath

from sherlockml.datasets import session


def rationalise_projectpath(path):

    # All paths should be relative to root
    path = posixpath.join("/", path)

    normed = posixpath.normpath(path)

    if path.endswith("/") and not normed.endswith("/"):
        normed += "/"

    return normed


def projectpath_to_bucketpath(project_path, project_id=None):

    if project_id is None:
        project_id = session.project_id_from_environment()

    # Project path will already start with '/' so just prepend the project ID
    bucket_path = project_id + rationalise_projectpath(project_path)

    return bucket_path


def bucketpath_to_projectpath(path):
    """Drop the project ID from the front of the path."""
    parts = path.split("/")
    return posixpath.join("/", *parts[1:])


def project_relative_path(project_root, project_path):

    project_root = rationalise_projectpath(project_root)
    project_path = rationalise_projectpath(project_path)

    if not project_path.startswith(project_root):
        tpl = "{} is not a sub path of {}"
        raise ValueError(tpl.format(project_path, project_root))

    # Remove the root
    relative_path = project_path[len(project_root) :]

    # Get rid of any leading '/'es
    relative_path = relative_path.lstrip("/")

    return relative_path


def project_parent_directories(project_path):
    """List all the directories in the tree containing this file.

    Parameters
    ----------
    project_path : str
        The file to list the parent directories of

    Returns
    -------
    list of str
        The paths of the parent directories
    """

    # Ensure in assumed format - can now assume to be absolute
    project_path = rationalise_projectpath(project_path)

    # Stripping trailing slashes as if it's a directory we still just want to
    # get its parent
    dirname = posixpath.dirname(project_path.rstrip("/"))

    directories = []

    parts = dirname.split("/")
    for i_last in range(1, len(parts) + 1):
        directories.append("/".join(parts[:i_last]) + "/")

    return directories



import os
from collections import namedtuple

from six.moves.configparser import ConfigParser, NoSectionError, NoOptionError


Profile = namedtuple(
    "Profile", ["domain", "protocol", "client_id", "client_secret"]
)

DEFAULT_PROFILE = "default"
DEFAULT_DOMAIN = "services.sherlockml.com"
DEFAULT_PROTOCOL = "https"


def load(path):
    """Read the SherlockML configuration from a file."""

    parser = ConfigParser()
    parser.read(str(path))

    def _get(section, option):
        try:
            return parser.get(section, option)
        except (NoSectionError, NoOptionError):
            return None

    profiles = {}

    for section in parser.sections():
        profiles[section] = Profile(
            domain=_get(section, "domain"),
            protocol=_get(section, "protocol"),
            client_id=_get(section, "client_id"),
            client_secret=_get(section, "client_secret"),
        )

    return profiles


def load_profile(path, profile):
    """Read a SherlockML profile from a file."""
    profiles = load(path)
    try:
        return profiles[profile]
    except KeyError:
        return Profile(None, None, None, None)


def _default_credentials_path():
    xdg_config_home = os.environ.get("XDG_CONFIG_HOME")
    if not xdg_config_home:
        xdg_config_home = os.path.expanduser("~/.config")
    return os.path.join(xdg_config_home, "sherlockml", "credentials")


class CredentialsError(RuntimeError):
    pass


def _raise_credentials_error(type_):
    raise CredentialsError("No {} found".format(type_))


def resolve_profile(
    credentials_path=None,
    profile_name=None,
    domain=None,
    protocol=None,
    client_id=None,
    client_secret=None,
):

    resolved_credentials_path = (
        credentials_path
        or os.getenv("SHERLOCKML_CREDENTIALS_PATH")
        or _default_credentials_path()
    )

    resolved_profile_name = (
        profile_name or os.getenv("SHERLOCKML_PROFILE") or DEFAULT_PROFILE
    )

    profile = load_profile(resolved_credentials_path, resolved_profile_name)

    resolved_domain = (
        domain
        or os.getenv("SHERLOCKML_DOMAIN")
        or profile.domain
        or DEFAULT_DOMAIN
    )

    resolved_protocol = (
        protocol
        or os.getenv("SHERLOCKML_PROTOCOL")
        or profile.protocol
        or DEFAULT_PROTOCOL
    )

    resolved_client_id = (
        client_id
        or os.getenv("SHERLOCKML_CLIENT_ID")
        or profile.client_id
        or _raise_credentials_error("client_id")
    )

    resolved_client_secret = (
        client_secret
        or os.getenv("SHERLOCKML_CLIENT_SECRET")
        or profile.client_secret
        or _raise_credentials_error("client_secret")
    )

    return Profile(
        domain=resolved_domain,
        protocol=resolved_protocol,
        client_id=resolved_client_id,
        client_secret=resolved_client_secret,
    )



# Copyright 2018 ASI Data Science
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum
from collections import namedtuple

from marshmallow import (
    Schema,
    fields,
    post_load,
    validates_schema,
    ValidationError,
)
from marshmallow_enum import EnumField

from sherlockml.clients.base import BaseClient


class FileNodeType(Enum):
    FILE = "file"
    DIRECTORY = "directory"


File = namedtuple("File", ["path", "name", "last_modified", "size"])
Directory = namedtuple(
    "Directory",
    ["path", "name", "last_modified", "size", "truncated", "content"],
)
ListResponse = namedtuple("ListResponse", ["project_id", "path", "content"])


class FileNodeSchema(Schema):

    path = fields.String(required=True)
    name = fields.String(required=True)
    type = EnumField(FileNodeType, by_value=True, required=True)
    last_modified = fields.DateTime(required=True)
    size = fields.Integer(required=True)
    truncated = fields.Boolean()
    content = fields.Nested("self", many=True)

    @validates_schema
    def validate_type(self, data):
        if data["type"] == FileNodeType.DIRECTORY:
            required_fields = Directory._fields
        elif data["type"] == FileNodeType.FILE:
            required_fields = File._fields
        if set(data.keys()) != set(required_fields).union({"type"}):
            raise ValidationError("Wrong fields for {}.".format(data["type"]))

    @post_load
    def make_file_node(self, data):
        if data["type"] == FileNodeType.DIRECTORY:
            return Directory(**{key: data[key] for key in Directory._fields})
        elif data["type"] == FileNodeType.FILE:
            return File(**{key: data[key] for key in File._fields})
        else:
            raise ValueError("Invalid file node type.")


class ListResponseSchema(Schema):

    project_id = fields.UUID(data_key="project_id", required=True)
    path = fields.String(required=True)
    content = fields.List(fields.Nested(FileNodeSchema), required=True)

    @post_load
    def make_list_response(self, data):
        return ListResponse(**data)


class WorkspaceClient(BaseClient):

    SERVICE_NAME = "workspace"

    def list(self, project_id, prefix, depth):
        endpoint = "/project/{}/file".format(project_id)
        params = {"depth": depth, "prefix": prefix}
        response = self._get(endpoint, ListResponseSchema(), params=params)
        return response.content



# Copyright 2018 ASI Data Science
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from decimal import Decimal

import pytest
from marshmallow import ValidationError

from sherlockml.clients.cluster import NodeType, NodeTypeSchema, ClusterClient
from tests.clients.fixtures import PROFILE


NODE_TYPE = NodeType(
    id="test-id",
    name="test-name",
    instance_group="test-instance-group",
    max_interactive_instances=10,
    max_job_instances=0,
    milli_cpus=1000,
    memory_mb=4096,
    num_gpus=0,
    gpu_name=None,
    cost_usd_per_hour=Decimal("4.2"),
)

NODE_TYPE_BODY = {
    "nodeTypeId": "test-id",
    "name": "test-name",
    "instanceGroup": "test-instance-group",
    "milliCpus": 1000,
    "memoryMb": 4096,
    "numGpus": 0,
    "gpuName": None,
    "costUsdPerHour": 4.2,
    "maxInteractiveInstances": 10,
    "maxJobInstances": 0,
}

NODE_TYPE_DEFAULT = NodeType(
    id="other-test-id",
    name=None,
    instance_group=None,
    max_interactive_instances=0,
    max_job_instances=0,
    milli_cpus=1000,
    memory_mb=4096,
    num_gpus=0,
    gpu_name=None,
    cost_usd_per_hour=Decimal("4.2"),
)

NODE_TYPE_BODY_DEFAULT = {
    "nodeTypeId": "other-test-id",
    "name": None,
    "instanceGroup": None,
    "milliCpus": 1000,
    "memoryMb": 4096,
    "numGpus": 0,
    "gpuName": None,
    "costUsdPerHour": 4.2,
    "maxInteractiveInstances": 0,
    "maxJobInstances": 0,
}


def test_node_type_schema():
    data = NodeTypeSchema().load(NODE_TYPE_BODY)
    assert data == NODE_TYPE


def test_node_type_schema_with_defaults():
    data = NodeTypeSchema().load(NODE_TYPE_BODY_DEFAULT)
    assert data == NODE_TYPE_DEFAULT


def test_node_type_schema_load_invalid():
    with pytest.raises(ValidationError):
        NodeTypeSchema().load({})


@pytest.mark.parametrize(
    "kwargs, query_params",
    [
        ({}, {}),
        (
            {"interactive_instances_configured": True},
            {"interactiveInstancesConfigured": "true"},
        ),
        (
            {"interactive_instances_configured": False},
            {"interactiveInstancesConfigured": "false"},
        ),
        (
            {"job_instances_configured": True},
            {"jobInstancesConfigured": "true"},
        ),
        (
            {"job_instances_configured": False},
            {"jobInstancesConfigured": "false"},
        ),
        (
            {
                "interactive_instances_configured": True,
                "job_instances_configured": True,
            },
            {
                "interactiveInstancesConfigured": "true",
                "jobInstancesConfigured": "true",
            },
        ),
        (
            {
                "interactive_instances_configured": True,
                "job_instances_configured": False,
            },
            {
                "interactiveInstancesConfigured": "true",
                "jobInstancesConfigured": "false",
            },
        ),
    ],
)
def test_cluster_client_list_single_tenanted_node_types(
    mocker, kwargs, query_params
):
    mocker.patch.object(ClusterClient, "_get", return_value=[NODE_TYPE])
    schema_mock = mocker.patch("sherlockml.clients.cluster.NodeTypeSchema")

    client = ClusterClient(PROFILE)
    assert client.list_single_tenanted_node_types(**kwargs) == [NODE_TYPE]

    schema_mock.assert_called_once_with(many=True)
    ClusterClient._get.assert_called_once_with(
        "/node-type/single-tenanted",
        schema_mock.return_value,
        params=query_params,
    )


def test_cluster_client_configure_single_tenanted_node_type(mocker):
    put_raw_mock = mocker.patch.object(ClusterClient, "_put_raw")

    client = ClusterClient(PROFILE)
    response = client.configure_single_tenanted_node_type(
        NODE_TYPE.id,
        NODE_TYPE.name,
        NODE_TYPE.instance_group,
        NODE_TYPE.max_interactive_instances,
        NODE_TYPE.max_job_instances,
    )

    assert response == put_raw_mock.return_value

    ClusterClient._put_raw.assert_called_once_with(
        "/node-type/single-tenanted/{}/configuration".format(NODE_TYPE.id),
        json={
            "name": NODE_TYPE.name,
            "instanceGroup": NODE_TYPE.instance_group,
            "maxInteractiveInstances": NODE_TYPE.max_interactive_instances,
            "maxJobInstances": NODE_TYPE.max_job_instances,
        },
    )


def test_cluster_client_disable_single_tenanted_node_type(mocker):
    delete_raw_mock = mocker.patch.object(ClusterClient, "_delete_raw")

    client = ClusterClient(PROFILE)
    response = client.disable_single_tenanted_node_type(NODE_TYPE.id)

    assert response == delete_raw_mock.return_value

    ClusterClient._delete_raw.assert_called_once_with(
        "/node-type/single-tenanted/{}/configuration".format(NODE_TYPE.id)
    )



# Copyright 2018 ASI Data Science
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import fnmatch
import os
import posixpath
import contextlib
import tempfile
import io

from botocore.client import ClientError

from sherlockml.datasets import path, session
from sherlockml.datasets.session import SherlockMLDatasetsError


# For backwards compatability
SherlockMLFileSystemError = SherlockMLDatasetsError


def _s3_client(project_id=None):
    """Generate a boto S3 client to access this project's datasets."""

    # At present, calls to the interface of this module run this function once
    # to generate an S3 client, then pass it around to avoid making many calls
    # to the secret service. This could result in an invalid client in the
    # middle of an operation, if we get unlucky and the AWS credentials expire
    # at that moment.

    # In the future, we may wish to make a wrapper class that manages the S3
    # client and proxies calls to its methods. This wrapper would catch
    # failures resulting from invalid AWS credentials and build a new client
    # and retry the operation accordingly.

    return session.get().s3_client(project_id)


def _bucket(project_id=None):
    return session.get().bucket(project_id)


def ls(prefix="/", project_id=None, show_hidden=False, s3_client=None):
    """List contents of project datasets.

    Parameters
    ----------
    prefix : str, optional
        List only files in the datasets matching this prefix. Default behaviour
        is to list all files.
    project_id : str, optional
        The project to list files from. You need to have access to this project
        for it to work. Defaults to the project set by SHERLOCKML_PROJECT_ID in
        your environment.
    show_hidden : bool, optional
        Include hidden files in the output. Defaults to False.
    s3_client : botocore.client.S3, optional
        Advanced - a specific boto client for AWS S3 to use.

    Returns
    -------
    list
        The list of files from the project datasets.
    """

    if s3_client is None:
        s3_client = _s3_client(project_id)

    bucket = _bucket(project_id)

    # Use a paginator to enable listing projects with more than 1000 files
    paginator = s3_client.get_paginator("list_objects_v2")
    response_iterator = paginator.paginate(
        Bucket=bucket,
        Prefix=path.projectpath_to_bucketpath(prefix, project_id),
    )

    paths = []
    for part in response_iterator:

        try:
            objects = part["Contents"]
        except KeyError:
            continue

        for obj in objects:
            project_path = path.bucketpath_to_projectpath(obj["Key"])

            # Ignore the root of the project directory
            if project_path != "/":
                paths.append(project_path)

    if show_hidden:
        return paths
    else:
        non_hidden_paths = [
            path_
            for path_ in paths
            if not any(element.startswith(".") for element in path_.split("/"))
        ]
        return non_hidden_paths


def glob(
    pattern, prefix="/", project_id=None, show_hidden=False, s3_client=None
):
    """List contents of project datasets that match a glob pattern.

    Parameters
    ----------
    pattern : str
        The pattern that contents need to match.
    prefix : str, optional
        List only files in the project datasets that have this prefix. Default
        behaviour is to list all files.
    project_id : str, optional
        The project to list files from. You need to have access to this project
        for it to work. Defaults to the project set by SHERLOCK_PROJECT_ID in
        your environment.
    show_hidden : bool, optional
        Include hidden files in the output. Defaults to False.
    s3_client : botocore.client.S3, optional
        Advanced - a specific boto client for AWS S3 to use.

    Returns
    -------
    list
        The list of files from the project that match the glob pattern.
    """

    contents = ls(
        prefix=prefix,
        project_id=project_id,
        show_hidden=show_hidden,
        s3_client=s3_client,
    )

    return fnmatch.filter(contents, pattern)


def _isdir(project_path, project_id=None, s3_client=None):
    """Determine if a path in a project's datasets is a directory.

    Parameters
    ----------
    project_path : str
        The path in the project datasets to test.
    project_id : str, optional
        The project to list files from. You need to have access to this project
        for it to work. Defaults to the project set by SHERLOCKML_PROJECT_ID in
        your environment.
    s3_client : botocore.client.S3, optional
        Advanced - a specific boto client for AWS S3 to use.

    Returns
    -------
    bool
    """
    # 'Directories' in the S3 bucket always end in a '/'
    if not project_path.endswith("/"):
        project_path += "/"
    matches = ls(
        project_path,
        project_id=project_id,
        show_hidden=True,
        s3_client=s3_client,
    )
    return len(matches) >= 1


def _isfile(project_path, project_id=None, s3_client=None):
    """Determine if a path in a project's datasets is a file.

    Parameters
    ----------
    project_path : str
        The path in the project directory to test.
    project_id : str, optional
        The project to list files from. You need to have access to this project
        for it to work. Defaults to the project set by SHERLOCKML_PROJECT_ID in
        your environment.
    s3_client : botocore.client.S3, optional
        Advanced - a specific boto client for AWS S3 to use.

    Returns
    -------
    bool
    """
    if _isdir(project_path, project_id, s3_client):
        return False
    matches = ls(
        project_path,
        project_id=project_id,
        show_hidden=True,
        s3_client=s3_client,
    )
    rationalised_path = path.rationalise_projectpath(project_path)
    return any(match == rationalised_path for match in matches)


def _create_parent_directories(project_path, project_id, s3_client):

    bucket = _bucket(project_id)

    # Make sure empty objects exist for directories
    # List once for speed
    all_objects = set(
        ls("/", project_id=project_id, show_hidden=True, s3_client=s3_client)
    )

    for dirname in path.project_parent_directories(project_path):

        if dirname == "/":
            # Root is not returned by ls
            continue

        # We're doing this manually instead of using _isdir as _isdir will
        # return true if '/somedir/myfile' exists, even if '/somedir/' does not
        if dirname not in all_objects:
            bucket_path = path.projectpath_to_bucketpath(dirname, project_id)
            # Directories on S3 are empty objects with trailing '/' on the key
            s3_client.put_object(
                Bucket=bucket, Key=bucket_path, ServerSideEncryption="AES256"
            )


def _put_file(local_path, project_path, project_id, s3_client):

    bucket = _bucket(project_id)
    bucket_path = path.projectpath_to_bucketpath(project_path, project_id)

    if bucket_path.endswith("/"):
        msg = (
            "the destination path {} indicates a directory but the "
            "source path {} is a normal file - please provide a full "
            "destination path"
        ).format(repr(project_path), repr(local_path))
        raise SherlockMLDatasetsError(msg)

    s3_client.upload_file(
        local_path,
        bucket,
        bucket_path,
        ExtraArgs={"ServerSideEncryption": "AES256"},
    )


def _put_directory(local_path, project_path, project_id, s3_client):

    bucket = _bucket(project_id)
    bucket_path = path.projectpath_to_bucketpath(project_path, project_id)

    # Directories on S3 are empty objects with trailing '/' on the key
    if not bucket_path.endswith("/"):
        bucket_path += "/"
    s3_client.put_object(
        Bucket=bucket, Key=bucket_path, ServerSideEncryption="AES256"
    )

    # Recursively put the contents of the directory
    for entry in os.listdir(local_path):
        _put_recursive(
            os.path.join(local_path, entry),
            posixpath.join(project_path, entry),
            project_id,
            s3_client,
        )


def _put_recursive(local_path, project_path, project_id, s3_client):
    """Puts a file/directory without checking that parent directory exists."""
    if os.path.isdir(local_path):
        _put_directory(local_path, project_path, project_id, s3_client)
    else:
        _put_file(local_path, project_path, project_id, s3_client)


def put(local_path, project_path, project_id=None):
    """Copy from the local filesystem to a project's datasets.

    Parameters
    ----------
    local_path : str or os.PathLike
        The source path in the local filesystem to copy.
    project_path : str
        The destination path in the project directory.
    project_id : str, optional
        The project to put files in. You need to have access to this project
        for it to work. Defaults to the project set by SHERLOCKML_PROJECT_ID in
        your environment.
    """

    if hasattr(os, "fspath"):
        local_path = os.fspath(local_path)

    s3_client = _s3_client(project_id)

    _create_parent_directories(project_path, project_id, s3_client)
    _put_recursive(local_path, project_path, project_id, s3_client)


def _get_file(project_path, local_path, project_id, s3_client):

    if local_path.endswith("/"):
        msg = (
            "the destination path {} indicates a directory but the "
            "source path {} is a normal file - please provide a full "
            "destination path"
        ).format(repr(project_path), repr(local_path))
        raise SherlockMLDatasetsError(msg)

    bucket = _bucket(project_id)
    bucket_path = path.projectpath_to_bucketpath(project_path, project_id)

    try:
        s3_client.download_file(bucket, bucket_path, local_path)
    except ClientError as err:
        if "404" in err.args[0]:
            msg = "no file {} in project".format(repr(project_path))
            raise SherlockMLDatasetsError(msg)
        else:
            raise


def _get_directory(project_path, local_path, project_id, s3_client):

    # Firstly, make sure that the location to write to locally exists
    containing_dir = os.path.dirname(local_path)
    if not containing_dir:
        containing_dir = "."
    if not os.path.isdir(containing_dir):
        msg = "No such directory: {}".format(repr(containing_dir))
        raise IOError(msg)

    paths_to_get = ls(
        project_path,
        project_id=project_id,
        show_hidden=True,
        s3_client=s3_client,
    )
    for object_path in paths_to_get:

        local_dest = os.path.join(
            local_path, path.project_relative_path(project_path, object_path)
        )

        if object_path.endswith("/"):
            # Objects with a trailing '/' on S3 indicate directories
            if not os.path.exists(local_dest):
                os.makedirs(local_dest)
        else:
            # Make sure directory exists to put files into
            dirname = os.path.dirname(local_dest)
            if not os.path.exists(dirname):
                os.makedirs(dirname)
            _get_file(object_path, local_dest, project_id, s3_client)


def get(project_path, local_path, project_id=None):
    """Copy from a project's datasets to the local filesystem.

    Parameters
    ----------
    project_path : str
        The source path in the project datasets to copy.
    local_path : str or os.PathLike
        The destination path in the local filesystem.
    project_id : str, optional
        The project to get files from. You need to have access to this project
        for it to work. Defaults to the project set by SHERLOCKML_PROJECT_ID in
        your environment.
    """

    if hasattr(os, "fspath"):
        local_path = os.fspath(local_path)

    client = _s3_client(project_id)

    if _isdir(project_path, project_id, client):
        _get_directory(project_path, local_path, project_id, client)
    else:
        _get_file(project_path, local_path, project_id, client)


def mv(source_path, destination_path, project_id=None):
    """Move a file within a project's datasets.

    Parameters
    ----------
    source_path : str
        The source path in the project datasets to move.
    destination_path : str
        The destination path in the project datasets.
    project_id : str, optional
        The project to get files from. You need to have access to this project
        for it to work. Defaults to the project set by SHERLOCKML_PROJECT_ID in
        your environment.
    """

    s3_client = _s3_client(project_id)

    cp(source_path, destination_path, project_id, s3_client)
    rm(source_path, project_id, s3_client)


def cp(source_path, destination_path, project_id=None, s3_client=None):
    """Copy a file within a project's datasets.

    Parameters
    ----------
    source_path : str
        The source path in the project datasets to copy.
    destination_path : str
        The destination path in the project datasets.
    project_id : str, optional
        The project to get files from. You need to have access to this project
        for it to work. Defaults to the project set by SHERLOCKML_PROJECT_ID in
        your environment.
    s3_client : botocore.client.S3, optional
        Advanced - a specific boto client for AWS S3 to use.
    """

    if s3_client is None:
        s3_client = _s3_client(project_id)

    if not _isfile(source_path, project_id, s3_client):
        raise SherlockMLDatasetsError("source_path must be a file")

    if destination_path.endswith("/"):
        raise SherlockMLDatasetsError("destination_path must be a file path")

    bucket = _bucket(project_id)
    source_bucket_path = path.projectpath_to_bucketpath(
        source_path, project_id
    )
    destination_bucket_path = path.projectpath_to_bucketpath(
        destination_path, project_id
    )

    copy_source = {"Bucket": bucket, "Key": source_bucket_path}
    s3_client.copy(
        copy_source,
        bucket,
        destination_bucket_path,
        ExtraArgs={"ServerSideEncryption": "AES256"},
    )


def rm(project_path, project_id=None, s3_client=None):
    """Remove a file from the project directory.

    Parameters
    ----------
    project_path : str
        The path in the project datasets to remove.
    project_id : str, optional
        The project to get files from. You need to have access to this project
        for it to work. Defaults to the project set by SHERLOCKML_PROJECT_ID in
        your environment.
    s3_client : botocore.client.S3, optional
        Advanced - a specific boto client for AWS S3 to use.
    """

    if s3_client is None:
        s3_client = _s3_client(project_id)

    if not _isfile(project_path, project_id, s3_client):
        raise SherlockMLDatasetsError("not a file")

    bucket = _bucket(project_id)
    bucket_path = path.projectpath_to_bucketpath(project_path, project_id)

    s3_client.delete_object(Bucket=bucket, Key=bucket_path)


def rmdir(project_path, project_id=None):
    """Remove a directory from the project datasets.

    Parameters
    ----------
    remote_path : str
        The path of the directory to remove.
    project_id : str, optional
        The project to get files from. You need to have access to this project
        for it to work. Defaults to the project set by SHERLOCKML_PROJECT_ID in
        your environment.
    """

    s3_client = _s3_client(project_id)

    if not _isdir(project_path, project_id, s3_client):
        raise SherlockMLDatasetsError("not a directory")

    contents = ls(
        project_path, project_id, show_hidden=True, s3_client=s3_client
    )
    if not len(contents) == 1:
        raise SherlockMLDatasetsError("directory is not empty")

    # Directory paths must end with '/'
    if not project_path.endswith("/"):
        project_path += "/"

    bucket = _bucket(project_id)
    bucket_path = path.projectpath_to_bucketpath(project_path, project_id)

    s3_client.delete_object(Bucket=bucket, Key=bucket_path)


def etag(project_path, project_id=None):
    """Get a unique identifier for the current version of a file.

    Parameters
    ----------
    project_path : str
        The path in the project datasets.
    project_id : str, optional
        The project to get files from. You need to have access to this project
        for it to work. Defaults to the project set by SHERLOCKML_PROJECT_ID in
        your environment.

    Returns
    -------
    str
    """

    client = _s3_client(project_id)

    bucket = _bucket(project_id)
    bucket_path = path.projectpath_to_bucketpath(project_path, project_id)

    s3_object = client.get_object(Bucket=bucket, Key=bucket_path)

    return s3_object["ETag"].strip('"')


@contextlib.contextmanager
def open(project_path, mode="r", temp_dir=None, **kwargs):
    """Open a file from a project's datasets for reading.

    This downloads the file into a temporary directory before opening it, so if
    your files are very large, this function can take a long time.

    Parameters
    ----------
    project_path : str
        The path of the file in the project's datasets to open.
    mode : str
        The opening mode, either 'r' or 'rb'. This is passed down to the
        standard python open function. Writing is currently not supported.
    temp_dir : str
        A directory on the local filesystem where you would like the file to be
        saved into temporarily. Note that on SherlockML servers, the default
        temporary directory can break with large files, so if your file is
        upwards of 2GB, it is recommended to specify temp_dir='/project'.
    """

    if _isdir(project_path):
        raise SherlockMLDatasetsError("Can't open directories.")

    if any(char in mode for char in ("w", "a", "x")):
        raise NotImplementedError("Currently, only reading is implemented.")

    tmpdir = tempfile.mkdtemp(prefix=".", dir=temp_dir)
    local_path = os.path.join(tmpdir, os.path.basename(project_path))

    try:
        local_path = os.path.join(tmpdir, os.path.basename(project_path))
        get(project_path, local_path)
        with io.open(local_path, mode, **kwargs) as file_object:
            yield file_object
    finally:
        if os.path.isfile(local_path):
            os.remove(local_path)
        if os.path.isdir(tmpdir):
            os.rmdir(tmpdir)



# Copyright 2018 ASI Data Science
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from setuptools import setup, find_packages


setup(
    name="sherlockml",
    description="Python library for interacting with SherlockML.",
    url="https://sherlockml.com",
    author="ASI Data Science",
    author_email="engineering@asidatascience.com",
    license="Apache Software License",
    packages=find_packages(),
    use_scm_version={"version_scheme": "post-release"},
    setup_requires=["setuptools_scm", "pytest-runner"],
    tests_require=[
        "pytest",
        "pytest-mock",
        "requests_mock",
        "python-dateutil",
    ],
    install_requires=[
        "requests",
        "pytz",
        "six",
        "enum34; python_version<'3.4'",
        "marshmallow==3.0.0b12",
        "marshmallow_enum",
        "boto3",
        "botocore",
    ],
    dependency_links=[
        "git+https://github.com/marshmallow-code/marshmallow"
        "@3.0.0b12#egg=marshmallow"
    ],
)



# Copyright 2018 ASI Data Science
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import fnmatch
import os
import posixpath
import pytest

from sherlockml import datasets

from tests.datasets.fixtures import (  # noqa: F401
    mock_secret_client,
    project_env,
    project_directory,
    write_remote_object,
    read_remote_object,
    remote_file,
    remote_tree,
    local_file,
    local_file_changed,
    local_tree,
    temporary_directory,
    TEST_DIRECTORY,
    TEST_FILE_NAME,
    TEST_TREE,
    TEST_TREE_NO_HIDDEN_FILES,
    VALID_ROOT_PATHS,
    VALID_DIRECTORIES,
    VALID_FILES,
    INVALID_PATHS,
    TEST_FILE_CONTENT,
    TEST_NON_EXISTENT,
    EMPTY_DIRECTORY,
)


pytestmark = pytest.mark.usefixtures("project_directory")


@pytest.mark.parametrize(  # noqa: F811
    "path, show_hidden, expected",
    [(path, True, TEST_TREE) for path in VALID_ROOT_PATHS]
    + [(path, False, TEST_TREE_NO_HIDDEN_FILES) for path in VALID_ROOT_PATHS],
)
def test_ls_root(remote_tree, path, show_hidden, expected):
    assert set(datasets.ls(path, show_hidden=show_hidden)) == {
        "/" + path for path in expected
    }


@pytest.mark.parametrize(  # noqa: F811
    "prefix, show_hidden, expected",
    [(directory, True, TEST_TREE) for directory in VALID_DIRECTORIES]
    + [
        (directory, False, TEST_TREE_NO_HIDDEN_FILES)
        for directory in VALID_DIRECTORIES
    ],
)
def test_ls_subdirectory(remote_tree, prefix, show_hidden, expected):
    if prefix.startswith("./"):
        prefix = prefix[2:]
    prefix = prefix.lstrip("/")
    matches = ["/" + path for path in expected if path.startswith(prefix)]
    assert set(datasets.ls(prefix, show_hidden=show_hidden)) == set(matches)


@pytest.mark.parametrize(  # noqa: F811
    "pattern, prefix, show_hidden, expected",
    [("*dir2*", directory, True, TEST_TREE) for directory in VALID_DIRECTORIES]
    + [
        ("*file1", directory, False, TEST_TREE_NO_HIDDEN_FILES)
        for directory in VALID_DIRECTORIES
    ],
)
def test_glob(pattern, remote_tree, prefix, show_hidden, expected):
    if prefix.startswith("./"):
        prefix = prefix[2:]
    prefix = prefix.lstrip("/")
    matches = [
        "/" + path
        for path in expected
        if path.startswith(prefix) and fnmatch.fnmatch(path, pattern)
    ]
    assert set(datasets.glob(pattern, prefix, show_hidden=show_hidden)) == set(
        matches
    )


@pytest.mark.parametrize(  # noqa: F811
    "path, result",
    [(path, True) for path in VALID_DIRECTORIES]
    + [(path, False) for path in VALID_FILES + INVALID_PATHS],
)
def test_isdir(remote_tree, path, result):
    assert datasets._isdir(path) is result


@pytest.mark.parametrize(  # noqa: F811
    "path, result",
    [(path, True) for path in VALID_FILES]
    + [(path, False) for path in VALID_DIRECTORIES + INVALID_PATHS],
)
def test_isfile(remote_tree, path, result):
    assert datasets._isfile(path) is result


@pytest.mark.parametrize(  # noqa: F811
    "destination",
    [TEST_FILE_NAME, "/" + TEST_FILE_NAME, "./" + TEST_FILE_NAME],
)
def test_put_file(local_file, destination):
    datasets.put(local_file, destination)
    content = read_remote_object(TEST_FILE_NAME)
    assert content == TEST_FILE_CONTENT


def test_put_file_nonexistent_directory(local_file):  # noqa: F811
    datasets.put(local_file, "/path/to/newdir/test_file")
    for path in ["/path/", "/path/to/", "/path/to/newdir/"]:
        content = read_remote_object(path)
        assert content == b""
    content = read_remote_object("/path/to/newdir/test_file")
    assert content == TEST_FILE_CONTENT


@pytest.mark.parametrize(  # noqa: F811
    "destination",
    [
        "",
        "./",
        "/",
        TEST_DIRECTORY + "/",
        "./" + TEST_DIRECTORY + "/",
        "/" + TEST_DIRECTORY + "/",
    ],
)
def test_put_file_in_directory(local_file, destination):
    with pytest.raises(datasets.SherlockMLDatasetsError):
        datasets.put(local_file, destination)


@pytest.mark.parametrize(  # noqa: F811
    "destination, resolved_destination",
    [
        ("", ""),
        ("./", ""),
        ("/", ""),
        (TEST_DIRECTORY, TEST_DIRECTORY),
        ("./" + TEST_DIRECTORY, TEST_DIRECTORY),
        ("/" + TEST_DIRECTORY, TEST_DIRECTORY),
        (TEST_DIRECTORY + "/", TEST_DIRECTORY),
        ("./" + TEST_DIRECTORY + "/", TEST_DIRECTORY),
        ("/" + TEST_DIRECTORY + "/", TEST_DIRECTORY),
    ],
)
def test_put_tree(local_tree, destination, resolved_destination):
    datasets.put(local_tree, destination)
    for filename in TEST_TREE:
        path = posixpath.join(resolved_destination, filename)
        content = read_remote_object(path)
        if filename.endswith("/"):
            assert content == b""
        else:
            assert content == TEST_FILE_CONTENT


def test_get_file(remote_file):  # noqa: F811
    with temporary_directory() as dirname:
        path = os.path.join(dirname, TEST_FILE_NAME)
        datasets.get(remote_file, path)
        with open(path, "rb") as f:
            content = f.read()
    assert content == TEST_FILE_CONTENT


def test_get_file_in_directory(remote_file):  # noqa: F811
    with temporary_directory() as dirname:
        # Ensure dirname ends with exactly one '/'
        dirname = dirname.rstrip("/") + "/"
        with pytest.raises(datasets.SherlockMLDatasetsError):
            datasets.get(remote_file, dirname)


def test_get_file_bad_local_path(remote_file):  # noqa: F811
    with temporary_directory() as dirname:
        path = os.path.join(dirname, "directory/does/not/exist")
        with pytest.raises(IOError):
            datasets.get(remote_file, path)


def _validate_local_tree(root, tree):
    for path in tree:
        full_path = os.path.join(root, path)
        if full_path.endswith("/"):
            assert os.path.isdir(full_path)
        else:
            with open(full_path, "rb") as f:
                content = f.read()
            assert content == TEST_FILE_CONTENT


@pytest.mark.parametrize("path", VALID_ROOT_PATHS)  # noqa: F811
def test_get_tree(remote_tree, path):
    with temporary_directory() as dirname:
        datasets.get(path, dirname)
        _validate_local_tree(dirname, TEST_TREE)


@pytest.mark.parametrize("path", VALID_ROOT_PATHS)  # noqa: F811
def test_get_tree_in_directory(remote_tree, path):
    with temporary_directory() as dirname:
        # Ensure dirname ends with exactly one '/'
        dirname = dirname.rstrip("/") + "/"
        datasets.get(path, dirname)
        _validate_local_tree(dirname, TEST_TREE)


@pytest.mark.parametrize("path", VALID_DIRECTORIES)  # noqa: F811
def test_get_subtree(remote_tree, path):
    prefix = path.strip("/") + "/"
    subtree = [
        tree_path[len(prefix) :]
        for tree_path in TEST_TREE
        if tree_path.startswith(prefix)
    ]
    with temporary_directory() as dirname:
        datasets.get(path, dirname)
        _validate_local_tree(dirname, subtree)


@pytest.mark.parametrize("path", VALID_DIRECTORIES)  # noqa: F811
def test_get_subtree_in_directory(remote_tree, path):
    prefix = path.strip("/") + "/"
    subtree = [
        tree_path[len(prefix) :]
        for tree_path in TEST_TREE
        if tree_path.startswith(prefix)
    ]
    with temporary_directory() as dirname:
        # Ensure dirname ends with exactly one '/'
        dirname = dirname.rstrip("/") + "/"
        datasets.get(path, dirname)
        _validate_local_tree(dirname, subtree)


@pytest.mark.parametrize("path", VALID_DIRECTORIES)  # noqa: F811
def test_get_tree_bad_local_path(remote_tree, path):
    with temporary_directory() as dirname:
        local_path = os.path.join(dirname, "directory/does/not/exist/")
        with pytest.raises(IOError):
            datasets.get(path, local_path)


def test_get_non_existent():
    with temporary_directory() as dirname:
        with pytest.raises(datasets.SherlockMLDatasetsError):
            datasets.get(
                TEST_NON_EXISTENT, os.path.join(dirname, TEST_NON_EXISTENT)
            )


@pytest.mark.parametrize("destination", ["new_file"])  # noqa: F811
def test_mv(remote_file, destination):
    datasets.mv(remote_file, destination)
    content = read_remote_object(destination)
    assert "/" + destination in datasets.ls()
    assert "/" + remote_file not in datasets.ls()
    assert content == TEST_FILE_CONTENT


@pytest.mark.parametrize("remote_dir, destination", [("/input/", "/output/")])
def test_mv_directory_source(remote_dir, destination):
    with pytest.raises(datasets.SherlockMLDatasetsError):
        datasets.mv(remote_dir, destination)


@pytest.mark.parametrize("destination", ["/output/"])  # noqa: F811
def test_mv_directory_destination(remote_file, destination):
    with pytest.raises(datasets.SherlockMLDatasetsError):
        datasets.mv(remote_file, destination)


@pytest.mark.parametrize("destination", ["new_file"])  # noqa: F811
def test_cp(remote_file, destination):
    datasets.cp(remote_file, destination)
    destination_content = read_remote_object(destination)
    source_content = read_remote_object(remote_file)
    assert "/" + destination in datasets.ls()
    assert destination_content == source_content


@pytest.mark.parametrize(
    "remote_dir, destination",
    [("/input/no-such-file", "/output/new-file"), ("/input/", "/output/")],
)
def test_cp_directory_source(remote_dir, destination):
    with pytest.raises(datasets.SherlockMLDatasetsError):
        datasets.cp(remote_dir, destination)


@pytest.mark.parametrize("destination", ["/output/"])  # noqa: F811
def test_cp_directory_destination(remote_file, destination):
    with pytest.raises(datasets.SherlockMLDatasetsError):
        datasets.cp(remote_file, destination)


def test_rm(remote_file):  # noqa: F811
    assert "/" + remote_file in datasets.ls()
    datasets.rm(remote_file)
    assert "/" + remote_file not in datasets.ls()


@pytest.mark.parametrize("remote_dir", ["/input/", "/input/no-such-file"])
def test_rm_directory_source(remote_dir):
    with pytest.raises(datasets.SherlockMLDatasetsError):
        datasets.rm(remote_dir)


def test_rmdir(remote_tree):  # noqa: F811
    assert "/" + EMPTY_DIRECTORY in datasets.ls()
    datasets.rmdir(EMPTY_DIRECTORY)
    assert "/" + EMPTY_DIRECTORY not in datasets.ls()


def test_rmdir_missing():
    with pytest.raises(datasets.SherlockMLDatasetsError):
        datasets.rmdir("missing/directory")


def test_rmdir_filetype(remote_tree):  # noqa: F811
    with pytest.raises(datasets.SherlockMLDatasetsError):
        datasets.rmdir(VALID_FILES[0])


def test_rmdir_nonempty(remote_tree):  # noqa: F811
    with pytest.raises(datasets.SherlockMLDatasetsError):
        datasets.rmdir("input/")


def test_etag(remote_file):  # noqa: F811
    etag = datasets.etag(remote_file)
    assert isinstance(etag, str)
    assert len(etag) > 0


def test_etag_change(remote_file, local_file_changed):  # noqa: F811
    initial_etag = datasets.etag(remote_file)
    datasets.put(local_file_changed, remote_file)
    final_etag = datasets.etag(remote_file)
    assert final_etag != initial_etag


def test_open_read(remote_file):  # noqa: F811
    with datasets.open(remote_file, "rb") as fp:
        assert fp.read() == TEST_FILE_CONTENT


def test_open_defaultmode(remote_file):  # noqa: F811
    with datasets.open(remote_file) as fp:
        assert fp.read() == TEST_FILE_CONTENT.decode("utf-8")


def test_open_missing():
    with pytest.raises(datasets.SherlockMLDatasetsError):
        with datasets.open("missing/file", "r"):
            pass


def test_open_directory(remote_tree):  # noqa: F811
    with pytest.raises(datasets.SherlockMLDatasetsError):
        with datasets.open("/input", "r"):
            pass
